%{

/*
   keys.l -- PIKT lexer for the keys.conf file

   $Header: /var/cvs/PIKT/pikt/src/piktc/keys.l,v 1.7 2005/01/04 13:48:31 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002,2003,2004,2005 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"
#include "piktc.h"
#include "keys.tab.h"

char _keyslex_[] = "keyslex";

#define RETURN(token) keyslval.sval = mystrdup(yytext, _keyslex_); return token

void keyserror(const char *);

%}

W		[ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]

%option noyywrap

%%

("//"|"#")[^\n]*\n		{ lineno++; }

^{L}[-_\.0-9a-zA-Z]*		{ RETURN(_HOST); }

{D}+				{ keyslval.ival = atoi(yytext);
				  return _INT;
				}

\"[^\"]+\"			{ *(yytext+yyleng-1) = '\0';
                                  keyslval.sval = mystrdup(yytext+1, _keyslex_);
                                  return _STR;
				}

[^ \t\n]+			{ RETURN(_STR); }

{W}				{ ; }

\n				{ lineno++; }

%%

/* keys error handler
 */
void
keyserror (const char *message)
{

	logreturn(OUTPUT, LOG_ERR, ERRID_KEYS_PARSE_ERROR, __FILE__, __LINE__, "keysparse", FALSE, "keys.conf, at or near line %d: %s: %s", lineno, message, yytext);
	bailout = TRUE;
	return;
}
