/*
   rpcsvc.c -- PIKT rpc service (piktc_svc) routines

   $Header: /var/cvs/PIKT/pikt/src/piktc/rpcsvc.c,v 1.92 2005/01/04 13:48:31 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002,2003,2004,2005 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#define PORTMAP

#include "../lib/common.h"

#include <rpc/rpc.h>

#if defined(HAVE_NETCONFIG_H)
#include <netconfig.h>
#endif

#include <rpc/xdr.h>
#include <rpc/clnt.h>
#include "rpc.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <sys/utsname.h>
#include <dirent.h>

#if defined(HAVE_SYS_SYSTEMINFO_H)
	#include <sys/systeminfo.h>
#endif

#include "piktc.h"

#ifdef NO_RPC_SVC
#define GET_KEY_1_SVC			get_key_1
#define LIST_FILE_1_SVC			list_file_1
#define INSTALL_FILE_1_SVC		install_file_1
#define CKSUM_FILE_1_SVC		cksum_file_1
#define DELETE_FILE_1_SVC		delete_file_1
#define RESTORE_FILE_1_SVC		restore_file_1
#define EDIT_PIKTD_CONF_1_SVC		edit_piktd_conf_1
#define EXECUTE_COMMAND_1_SVC		execute_command_1
#define EXECUTE_COMMANDWAIT_1_SVC	execute_commandwait_1
#define EXECUTE_ALERT_1_SVC		execute_alert_1
#define SHOW_ALERT_STATUS_1_SVC		show_alert_status_1
#define KILL_PIKTD_1_SVC		kill_piktd_1
#define KILL_PIKTC_SVC_1_SVC		kill_piktc_svc_1
#define RESTART_PIKTD_1_SVC		restart_piktd_1
#define RESTART_PIKTC_SVC_1_SVC		restart_piktc_svc_1
#define PIKTC_SVC_PID_1_SVC		piktc_svc_pid_1
#define REGISTER_REQUEST_1_SVC		register_request_1
#define CALL_BACK_1_SVC			call_back_1
#else
#define GET_KEY_1_SVC			get_key_1_svc
#define LIST_FILE_1_SVC			list_file_1_svc
#define INSTALL_FILE_1_SVC		install_file_1_svc
#define CKSUM_FILE_1_SVC		cksum_file_1_svc
#define DELETE_FILE_1_SVC		delete_file_1_svc
#define RESTORE_FILE_1_SVC		restore_file_1_svc
#define EDIT_PIKTD_CONF_1_SVC		edit_piktd_conf_1_svc
#define EXECUTE_COMMAND_1_SVC		execute_command_1_svc
#define EXECUTE_COMMANDWAIT_1_SVC	execute_commandwait_1_svc
#define EXECUTE_ALERT_1_SVC		execute_alert_1_svc
#define SHOW_ALERT_STATUS_1_SVC		show_alert_status_1_svc
#define KILL_PIKTD_1_SVC		kill_piktd_1_svc
#define KILL_PIKTC_SVC_1_SVC		kill_piktc_svc_1_svc
#define RESTART_PIKTD_1_SVC		restart_piktd_1_svc
#define RESTART_PIKTC_SVC_1_SVC		restart_piktc_svc_1_svc
#define PIKTC_SVC_PID_1_SVC		piktc_svc_pid_1_svc
#define REGISTER_REQUEST_1_SVC		register_request_1_svc
#define CALL_BACK_1_SVC			call_back_1_svc
#endif

/* number of seconds to sleep when entering one of the rpcsvc.c functions;
   this is to permit easier correlation of time stamps on both master and
   slave log messages--but only if the time is in sync on both master and
   slave!
*/
#ifdef DEEPDEBUG
  #define DEBUG_TIME_TO_SLEEP 5
#else
  #define DEBUG_TIME_TO_SLEEP 0
#endif

CLIENT *clnt;

req reqserv;

static execres xres;
static cksumres csres;
static keyencres keres;

int endian_master = 0;	/* endianness of master initially unknown */

int chkrequest (char *, ritm *, rfil *, struct svc_req *, const char *);
char *genkey(void);
int enckey(char *, unsigned int, char **, unsigned int *);
int authrand(void);
void free_execcmdout(void);

/* for all below, return 0 on success, else ERROR, NONE, DENIED, ...;
   log all errors including errno info, if available, in the local piktc.log
 */

/* first, attempt to authenticate the requesting piktc client system;
   compare the client name and/or client address to the approved master
   and/or master address; if they match, grant authentication; in special
   cases, make further attempts to grant authentication

   second, if access is set to local, check to see that the client's network
   domain matches the local network domain; if they dont' match, deny
   authentication

   third, check to see that the requesting account is an authorized account;
   if not, deny authentication

   fourth, optionally check to see that the requesting group is an authorized
   group; if not, deny authentication

   fifth, optionally check to see if the ethernet address of the requesting
   piktc client system matches the approved piktmaster ethernet address

   sixth, see if the keysum returned by the master matches the keysum
   established by the earlier get_key_svc() call ("strong authentication")

   seventh, check the PIKT.conf service access parameters to see if this
   request is authorized; if not, deny the request

   eighth, and if failed authentication hasn't already denied the request,
   do a call_back, to see if this request is registered on the master;
   if not, deny the request

   return values:  0 (approved), UNAUTHORIZED, DENIED
 */
int
chkrequest (char *kysm, ritm *an, rfil *rf, struct svc_req *rqstp, const char *request)
{
	static char fnnam[] = "chkrequest";
	int res;
	struct sockaddr_in *sock_in;
	struct hostent *hp;
	char client_name[MAXIDENT+1];
	char *cp;
	char *client_domain;
	char *hap;
	unsigned long clntaddr;
	char *halp;
	char client_address[128];	/* dotted-decimal string; be generous
					   to accommodate IPv6/IPng
					 */
	struct authunix_parms *unix_cred;
	char *aup_client_name;
	int aup_client_uid;
	int aup_client_gid;
	char cmd[MAXCMD+1];
	char arpres[MAXLINE+1];
	FILE *arp;
	int authenticated;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering chkrequest(), request is %s", request);
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	res = 0;

	authenticated = FALSE;

	aup_client_name = "<<ERR>>";
	aup_client_uid = -1;
	aup_client_gid = -1;

	/* gather the client info */

	sock_in = svc_getcaller(rqstp->rq_xprt);

	/* get the client host's (i.e., the would-be control machine's)
	   fully qualified domain name
	 */
	if (! (hp = gethostbyaddr((char *) &sock_in->sin_addr, sizeof(sock_in->sin_addr), AF_INET))) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_REVERSE_DNS_LOOKUP_ERROR, __FILE__, __LINE__, fnnam, FALSE, "for client address (sock_in->sin_addr.s_addr) %x, gethostbyaddr() returns null, reverse DNS lookup unsuccessful", sock_in->sin_addr.s_addr);
		res = DENIED;
		return res;
	}
	strncpy(client_name, hp->h_name, MAXIDENT);
	client_name[MAXIDENT] = '\0';	/* ensure null termination */
	client_domain = (cp = strchr(client_name, '.')) ? cp+1 : cp;
	while (! client_domain && (hap = *(hp->h_aliases))) {
		strncpy(client_name, hap, MAXIDENT);
		client_name[MAXIDENT] = '\0';	/* ensure null termination */
		hp->h_aliases++;
		client_domain = (cp = strchr(client_name, '.')) ? cp+1 : cp;
	}

	/* convert client address to host byte order */
	clntaddr = ntohl(sock_in->sin_addr.s_addr);

	/* get the dotted decimal form of the client address */
	strcpy(client_address, inet_ntoa(sock_in->sin_addr));

	/* get authunix credentials, including client_name, client_uid &
	   client_gid
	 */
	unix_cred = (struct authunix_parms *) rqstp->rq_clntcred;
	aup_client_name = unix_cred->aup_machname;
	aup_client_uid = unix_cred->aup_uid;
	aup_client_gid = unix_cred->aup_gid;

	/* first, attempt to authenticate this client (i.e., master); */

	if (strcmp(request, "call_back")) {	/* if is not call_back */

		/* typical cases */
		if (isauth_by_master && isauth_by_master_address) {
			if (! strcmp(client_name, master)           &&
			    ! strcmp(client_address, master_address)  ) {
				authenticated = TRUE;
			}
		} else if (isauth_by_master) {
			if (! strcmp(client_name, master)) {
				authenticated = TRUE;
			}
		} else if (isauth_by_master_address) {
			if (! strcmp(client_address, master_address)) {
				authenticated = TRUE;
			}
		}

		/* special cases */
		if (! authenticated) {
			/* if is validation self test */
			if (istest) {
				if (hostmatch(client_name, master)) {
					authenticated = TRUE;
				}
			/* else the client_name is not a fully qualified
			   host name, so cannot match master (which must be
			   a fully qualified host name); we will make one
			   last attempt to authenticate this client by
			   matching addresses
			 */
			 } else if (! client_domain            &&
			            ! isauth_by_master_address   ) {
				hp = gethostbyname(master);
				while (! authenticated              &&
				       (halp = *(hp->h_addr_list))   ) {
					if (memcmp(halp, &sock_in->sin_addr, sizeof(sock_in->sin_addr)) == 0) {
						authenticated = TRUE;
					}
					hp->h_addr_list++;
				}
			}
		}

		#ifdef REALSECURE
		if (authenticated) {
			/* this aup_client_name check is somewhat redundant,
			   given the earlier check of client_name
			 */
			if (! aup_client_name || ! strlen(aup_client_name) || strncasecmp(aup_client_name, master, strlen(aup_client_name))) {
				authenticated = FALSE;
			}
		}
		#endif

	} else {	/* is a call_back */

		/* for now, accept as valid a callback from any host
		   (but note that we still check the domain below, if
		   access is set to 'local', also the uid, and possibly
		   also the gid as well); in the future, we might want
		   to check against the systems list in systems.cfg
		 */
		authenticated = TRUE;

	}

	/* second, check the domain access */

	if (! strcasecmp(accessmode, "local")) {
		/* restrict requests to within the local domain only,
		   possibly setting authenticated to FALSE;
		   note that this check is also done for get_key & call_back
		 */
		if (! strcasecmp(domain, "domain")) {
			if (! client_domain || (! strstr(client_domain, domainname) && ! strstr(domainname, client_domain))) {
				authenticated = FALSE;
			}
		} else {	/* there is an actual specified domain */
			if (! client_domain || strcasecmp(client_domain, domain)) {
				authenticated = FALSE;
			}
		}
	} /* else accessmode is "global", so don't check anything */

	/* third, check the client uid */

	/* note that the uid check is done also in the case of get_key &
	   call_back, so the uid in the master and slave PIKT.conf
	   have to match!
	 */
	if (   ! istest
	    && (strcmp(request, "get_key"))
	    && (strcmp(request, "call_back"))
	    && (strcmp(request, "piktc_svc_pid"))
	    && (strcmp(request, "register_request"))
	   ) {
		if (aup_client_uid != authuid) {
			authenticated = FALSE;
		}
	}

	/* fourth, optionally check the client gid (if the gid was set
	   in PIKT.conf
	 */

	/* note that the gid check is done also in the case of get_key &
	   call_back, so the gid in the master and slave PIKT.conf
	   have to match!
	 */
	if (   ! istest
	    && (strcmp(request, "get_key"))
	    && (strcmp(request, "call_back"))
	    && (strcmp(request, "piktc_svc_pid"))
	    && (strcmp(request, "register_request"))
	   ) {
		if (PIKT_conf_gid && (aup_client_gid != authgid)) {
			authenticated = FALSE;
		}
	}

	/* fifth, optional check the client ethernet address (if this
	   was set in PIKT.conf
	 */

	if (isauth_by_master_ethernet_address && master_ethernet_address) {
		sprintf(cmd, "%s %s | %s %s", arpcmd, master, egrepcmd, master);
		if ((arp = popen(cmd, "r")) == NULL) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_PROC_OPEN_ERROR, __FILE__, __LINE__, fnnam, FALSE, "popen() error");
			authenticated = FALSE;
		} else {
			if (fgets(arpres, MAXLINE+1, arp) == NULL) { 
				if (ferror(arp))
					logreturn(PIKTLOG, LOG_ERR, ERRID_FGETS_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fgets() error for %s", cmd);
			}
			pclose(arp);
			if ((cp = strchr(arpres, '\n')))
				*cp = '\0';
			if (! ethaddr_match(arpres, master_ethernet_address))
				authenticated = FALSE;
		}
	}

	if (isdebug) {		/* show client authentication stuff */

		/* request info */
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "%s request from uid %d, gid %d on client %s (address %s)", request, aup_client_uid, aup_client_gid, client_name, client_address);

		/* from PIKT.conf */
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "uid (PIKT.conf uid) is %d", authuid);
		if (PIKT_conf_gid)
			logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "gid (PIKT.conf gid) is %d", authgid);
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "master (PIKT.conf master) is %s", master ? master : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "master_address (PIKT.conf master_address) is %s", master_address ? master_address : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "master_ethernet_address (PIKT.conf master_ethernet_address) is %s", master_ethernet_address ? master_ethernet_address : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "domain (PIKT.conf domain) is %s", domain ? domain : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "access (PIKT.conf access) is %s", accessmode ? accessmode : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "authenticate_by_master (PIKT.conf authenticate_by_master) is %s", isauth_by_master ? "YES" : "NO");
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "authenticate_by_master_address (PIKT.conf authenticate_by_master_address) is %s", isauth_by_master_address ? "YES" : "NO");
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "authenticate_by_master_ethernet_address (PIKT.conf authenticate_by_master_ethernet_address) is %s", isauth_by_master_ethernet_address ? "YES" : "NO");

		/* from local host */
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "local hostname (hostname) is %s", hostname ? hostname : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "local domainname (domainname) is %s", domainname ? domainname : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "local keysum (keysum) is %s", keysum ? keysum : "unknown");
		if (isauth_by_master_ethernet_address && master_ethernet_address)
			logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "local arp piktmaster result (arpres) is %s", arpres ? arpres : "unknown");

		/* from client */
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "client uid (aup_client_uid) is %d", aup_client_uid);
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "client gid (aup_client_gid) is %d", aup_client_gid);
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "client name (client_name) is %s", client_name ? client_name : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "client name (aup_client_name) is %s", aup_client_name ? aup_client_name : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "client address (sock_in->sin_addr.s_addr) is %x", sock_in->sin_addr.s_addr);
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "client address (clntaddr) is %x", clntaddr);
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "client address (client_address) is %s", client_address ? client_address : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "client domain (client_domain) is %s", client_domain ? client_domain : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "client keysum (kysm) is %s", kysm ? kysm : "unknown");

	}

	if (! authenticated) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_SERVICE_REQUEST_UNAUTHENTICATED, __FILE__, __LINE__, fnnam, FALSE, "unauthenticated %s request from uid %d, gid %d on client %s (address %s)", request, aup_client_uid, aup_client_gid, client_name, client_address);
		res = DENIED;
		return res;
	}

	/* if the client authenticates to this point, approve all get_key,
	   call_back, piktc_svc_pid, and register_request requests
	 */
	if (   (! strcmp(request, "get_key"))
	    || (! strcmp(request, "call_back"))
	    || (! strcmp(request, "piktc_svc_pid"))
	    || (! strcmp(request, "register_request"))
	   )
		return res;	/* should be 0 */

	/* sixth, check the keysum */
	if (kysm && keysum) {
		if (strncmp(kysm, keysum, strlen(keysum))) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_SERVICE_REQUEST_UNAUTHENTICATED, __FILE__, __LINE__, fnnam, FALSE, "unauthenticated %s request from uid %d, gid %d on client %s (address %s), kysm is %s, keysum is %s", request, aup_client_uid, aup_client_gid, client_name, client_address, kysm, keysum);
			res = DENIED;
			return res;
		}
	} else {
		logreturn(PIKTLOG, LOG_ERR, ERRID_KEYSUM_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "kysm is %s, keysum is %s", kysm ? kysm : "unknown", keysum ? keysum : "unknown");
		res = ERROR;
		return res;
	}

	/* seventh, check the PIKT.conf service access parameters to see if
	   this request is authorized
	 */

	if (services != ALL_SERVICES) {  /* one or more services disabled */
	
		if (isdebug) {
			logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "request is %s, services is %x", request, services);
		}

		/* get_key is always authorized
		if (! strcmp(request, "get_key")) {
			;
		} else
		 */

		/* call_back is always authorized
		if (! strcmp(request, "get_key")) {
			;
		} else
		 */

		if (! strcmp(request, "execute_command")) {
			if (! (services & EXECUTE_COMMANDS_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "execute_alert")) {
			if (! (services & EXECUTE_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "install_file")) {
			if (! strcmp(rf->dir, ALTDIR) &&
			    ! strcmp(rf->ext, ".alt") &&
			    ! (services & INSTALL_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! strcmp(rf->dir, OBJDIR) &&
			           ! strcmp(rf->ext, ".obj") &&
			           ! (services & INSTALL_OBJECTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if ((! strcmp(rf->dir, PRGDIR) ||
			            ! strcmp(rf->ext, "NIL")) &&
			           ! (services & INSTALL_PROGRAMS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! *(rf->ext) &&
			           ! (services & INSTALL_FILES_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "edit_piktd_conf")) {
			if (((an->state == ADD) || (an->aux && *(an->aux))) &&
			    ! (services & ENABLE_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (((an->state == DELETE) || (an->aux && ! *(an->aux))) &&
			           ! (services & DISABLE_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "show_alert_status")) {
			if (! (services & SHOW_ALERTS_STATUS_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "list_file")) {
			if (! strcmp(rf->dir, ALTDIR) &&
			    ! strcmp(rf->ext, ".alt") &&
			    ! (services & LIST_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! strcmp(rf->dir, OBJDIR) &&
			           ! strcmp(rf->ext, ".obj") &&
			           ! (services & LIST_OBJECTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if ((! strcmp(rf->dir, PRGDIR) ||
			            ! strcmp(rf->ext, "NIL")) &&
			           ! (services & LIST_PROGRAMS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! *(rf->ext) &&
			           ! (services & LIST_FILES_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (strstr(request, "fetch_file")) {
			/* since we now use execute_command() to do a "cat
			   <file>" for file fetches, execute_commands_service
			   must be set to true
			 */
			if (! (services & EXECUTE_COMMANDS_SERVICE)) {
				res = UNAUTHORIZED;
			} else {
				if (! strcmp(request, "fetch_file_alerts") &&
				    ! (services & FETCH_ALERTS_SERVICE)) {
					res = UNAUTHORIZED;
				} else if (! strcmp(request, "fetch_file_objects") &&
				           ! (services & FETCH_OBJECTS_SERVICE)) {
					res = UNAUTHORIZED;
				} else if (! strcmp(request, "fetch_file_programs") &&
				           ! (services & FETCH_PROGRAMS_SERVICE)) {
					res = UNAUTHORIZED;
				} else if (! strcmp(request, "fetch_file_files") &&
				           ! (services & FETCH_FILES_SERVICE)) {
					res = UNAUTHORIZED;
				}
			}

		} else if (! strcmp(request, "cksum_file")) {
			if (! strcmp(rf->dir, ALTDIR) &&
			    ! strcmp(rf->ext, ".alt") &&
			    ! (services & CKSUM_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! strcmp(rf->dir, OBJDIR) &&
			           ! strcmp(rf->ext, ".obj") &&
			           ! (services & CKSUM_OBJECTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if ((! strcmp(rf->dir, PRGDIR) ||
			            ! strcmp(rf->ext, "NIL")) &&
			           ! (services & CKSUM_PROGRAMS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! *(rf->ext) &&
			           ! (services & CKSUM_FILES_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "delete_file")) {
			if (! strcmp(rf->dir, ALTDIR) &&
			    ! strcmp(rf->ext, ".alt") &&
			    ! (services & DELETE_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! strcmp(rf->dir, OBJDIR) &&
			           ! strcmp(rf->ext, ".obj") &&
			           ! (services & DELETE_OBJECTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if ((! strcmp(rf->dir, PRGDIR) ||
			            ! strcmp(rf->ext, "NIL")) &&
			           ! (services & DELETE_PROGRAMS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! *(rf->ext) &&
			           ! (services & DELETE_FILES_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "restore_file")) {
			if (! strcmp(rf->dir, ALTDIR) &&
			    ! strcmp(rf->ext, ".alt") &&
			    ! (services & INSTALL_ALERTS_SERVICE) &&
			    ! (services & DELETE_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! strcmp(rf->dir, OBJDIR) &&
			           ! strcmp(rf->ext, ".obj") &&
			           ! (services & INSTALL_OBJECTS_SERVICE) &&
			           ! (services & DELETE_OBJECTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if ((! strcmp(rf->dir, PRGDIR) ||
			            ! strcmp(rf->ext, "NIL")) &&
			           ! (services & INSTALL_PROGRAMS_SERVICE) &&
			           ! (services & DELETE_PROGRAMS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! *(rf->ext) &&
			           ! (services & INSTALL_FILES_SERVICE) &&
			           ! (services & DELETE_FILES_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "kill_piktd")) {
			if (! (services & KILL_PIKTD_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "kill_piktc_svc")) {
			if (! (services & KILL_PIKTC_SVC_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "restart_piktd")) {
			if (! (services & RESTART_PIKTD_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "restart_piktc_svc")) {
			if (! (services & RESTART_PIKTC_SVC_SERVICE)) {
				res = UNAUTHORIZED;
			}
		}

		if (res == UNAUTHORIZED) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_SERVICE_REQUEST_UNAUTHENTICATED, __FILE__, __LINE__, fnnam, FALSE, "unauthorized %s request from uid %d, gid %d on client %s (address %s)", request, aup_client_uid, aup_client_gid, client_name, client_address);
			return res;
		}
	}

	/* eighth, do the call_back, to see if this request is registered on
	   the master; note: we don't call_back a call_back request!
	 */

	/* if call_back has not been turned off, and if this request is not
	   a get_key or call_back request (we should have returned already
	   if so, but we doublecheck here to prevent loops--defensive
	   programming!), do a call_back
	 */
	if (IS_CALL_BACK && strcmp(request, "get_key") && strcmp(request, "call_back")) {

		if (ismaster) {
			/* in effect, we are doing here what we would do if
			   doing a call_back() except looking entirely locally
			 */
			if (hostserv && rqstserv && hostmatch(hostname, hostserv)) {
				if (! strcmp(request, rqstserv)) {
					res = 0;  /* should still be 0 */
				/* the enable_alerts special case */
				} else if (! strcmp(request, "list_file") &&
					   ! strcmp(rqstserv, "edit_piktd_conf")) {
					res = 0;  /* should still be 0 */
				} else {
					res = DENIED;
				}
			} else if (rqstserv && (statserv == EXECUTE) && (! strcmp(request, "execute_command") || ! strcmp(request, "execute_alert"))) {
				if (! strcmp(request, rqstserv) || ! *rqstserv) {
					res = 0;
				}
			} else {
				res = DENIED;
			}
			if (isdebug || (res == DENIED)) {
				logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "hostserv is %s, rqstserv is %s, statserv is %d, hostname is %s, request is %s", hostserv ? hostserv : nullstring, rqstserv ? rqstserv : nullstring, statserv, hostname, request);
			}
		}
		/* no call_back for systems where NO_RPC_SVC is defined */
		#ifndef NO_RPC_SVC
		else {
			if (strstr(request, "fetch_file"))
				res = call_back(hostname, "execute_command", FALSE);
			else
				res = call_back(hostname, request, FALSE);
		}
		#endif
	}	/* else nothing:  we don't call_back a call_back! */

	if (res == DENIED) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_SERVICE_REQUEST_UNAUTHENTICATED, __FILE__, __LINE__, fnnam, FALSE, "unauthenticated %s request from uid %d, gid %d on client %s (address %s)", request, aup_client_uid, aup_client_gid, client_name, client_address);
	} else if (strcmp(request, "call_back") || isverboselog) {
		logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "%s request from uid %d, gid %d on client %s (address %s)", request, aup_client_uid, aup_client_gid, client_name, client_address);
	}

	/* finally, return the result */

	return res;
}

/* register a service request (the request and the hostname that the
   request is sent to);
   just to be safe, always service a register_request, whether IS_CALL_BACK
   or not
 */
int *
REGISTER_REQUEST_1_SVC (rqst *rq, struct svc_req *rqstp)
{
	static char fnnam[] = "REGISTER_REQUEST_SVC";
	static int res;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering REGISTER_REQUEST_SVC(), request is %s", rq->request);
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((res = chkrequest(rq->keysum.txttype_val, (ritm *)NULL, (rfil *)NULL, rqstp, "register_request")) != 0)
		return &res;

	FREE(hostserv);
	FREE(rqstserv);
	hostserv = mystrdup(rq->hostname, fnnam);
	rqstserv = mystrdup(rq->request, fnnam);
	statserv = rq->state;

	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "hostserv is %s, rqstserv is %s, statserv is %d", hostserv, rqstserv, statserv);

	return &res;
}

#ifndef NO_RPC_SVC

/* compare the request and hostname of the calling back piktc_svc with
   the locally registered request and hostname; if they match, return okay;
   return values: 0 (approved), DENIED; (chkrequest() can return
   UNAUTHORIZED, but not in the case of call_back, which is always
   authorized
 */
int *
CALL_BACK_1_SVC (rqst *rq, struct svc_req *rqstp)
{
	static char fnnam[] = "CALL_BACK_SVC";
	static int res;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering CALL_BACK_SVC(), request is %s", rq->request);
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if (IS_CALL_BACK) {

		if ((res = chkrequest((char *)NULL, (ritm *)NULL, (rfil *)NULL, rqstp, "call_back")) != 0)
			return &res;

		/* when doing a 'piktc -X' (execute with no wait) to many
		   different clients, all of which are doing callbacks
		   asynchronously, the currently registered hostserv might
		   not match the hostname of the machine doing the callback,
		   so in this case only we relax the requirement that the
		   local hostserv and the rq->hostname match; a match is
		   still required for 'piktc -x' (execute with wait) and
		   all other services
		 */
		if (hostserv && rqstserv && hostmatch(rq->hostname, hostserv)) {
			if (! strcmp(rq->request, rqstserv)) {
				res = 0;
			/* the enable_alerts special case */
			} else if (! strcmp(rq->request, "list_file") &&
			           ! strcmp(rqstserv, "edit_piktd_conf")) {
				res = 0;
			} else {
				res = DENIED;
			}
		} else if (rqstserv && (statserv == EXECUTE) && (! strcmp(rq->request, "execute_command") || ! strcmp(rq->request, "execute_alert"))) {
			if (! strcmp(rq->request, rqstserv) || ! *rqstserv) {
				res = 0;
			}
		} else {
			res = DENIED;
		}

		if (res == DENIED) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_SERVICE_REQUEST_UNAUTHENTICATED, __FILE__, __LINE__, fnnam, FALSE, "unauthorized request, hostserv is %s, rqstserv is %s, statserv is %d, rq->hostname is %s, rq->request is %s", hostserv ? hostserv : nullstring, rqstserv ? rqstserv : nullstring, statserv, rq->hostname, rq->request);
		} else if (isdebug) {
			logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "hostserv is %s, rqstserv is %s, statserv is %d, rq->hostname is %s, rq->request is %s", hostserv ? hostserv : nullstring, rqstserv ? rqstserv : nullstring, statserv, rq->hostname, rq->request);
		}

	} else {

		res = 0;	/* if call_back checks have been turned off,
				   always return success
				 */
	}

	return &res;
}

#endif	/* #ifndef NO_RPC_SVC */

/* generate an authentication/session key, compute its checksum, return the
   encrypted key to the master piktc
 */
keyencres *
GET_KEY_1_SVC (rqst *rq, struct svc_req *rqstp)
{
	static char fnnam[] = "GET_KEY_SVC";
	unsigned int keyenclen;
	char *keysan;
	char *keyencsan;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering GET_KEY_SVC(), request is %s", rq->request);
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((keres.result = chkrequest((char *)NULL, (ritm *)NULL, (rfil *)NULL, rqstp, "get_key")) != 0)
		return &keres;

	if (rq->enctyp != data_encryption_type) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "data_encryption_type mismatch (%d on master, %d on slave)", rq->enctyp, data_encryption_type);
		keres.result = ERROR;
		return &keres;
	}
	/* redo the key determination if enckey() returns ERROR (because,
	   for example, bf_setkey() reports a "weak key"
	 */
	do {
		FREE(key);
		key = genkey();
		keylen = strlen(key);
		FREE(keysum);
		keysum = checksumstr(5, key);
		FREE(keyenc);
	} while ((auth_encryption_type != 0) && enckey(key, keylen, &keres.keyencres_u.keyenc.txttype_val, &keres.keyencres_u.keyenc.txttype_len) == ERROR);
	if (auth_encryption_type == 0) {
		keres.keyencres_u.keyenc.txttype_val = mystrdup(key, fnnam);
		keres.keyencres_u.keyenc.txttype_len = strlen(key);
	}
	endian_master = rq->state;
	/* if the endianness of the master differs from the endianness of the
	   slave, and if we are encrypting the auth key, bring the keyenc
	   into endian conformity before returning to the master
	 */
	if ((auth_encryption_type != 0) && (endian_master != endian))
		fix_endian(keres.keyencres_u.keyenc.txttype_val, keres.keyencres_u.keyenc.txttype_len);
	keyenc = keres.keyencres_u.keyenc.txttype_val;
	keyenclen = keres.keyencres_u.keyenc.txttype_len;
	if (isdebug) {
		keysan = sanitize(key, keylen);
		keyencsan = sanitize(keyenc, keyenclen);
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "key (keysan) is %s, keysum is %s, keyenc (keyencsan) is %s, keyenclen is %d, endian is %d, endian_master is %d", /* key */ keysan, keysum, /* keyenc */ keyencsan, keyenclen, endian, endian_master);
		free(keysan);
		free(keyencsan);
	}
	keres.result = 0;
	return &keres;
}

/* encrypt the auth/session key
 */
int
enckey (char *ky, unsigned int kylen, char ** ke, unsigned int *kelen)
{
	static char fnnam[] = "enckey";
	int res;

	res = 0;
	bf_setkey(&bc, pvtkey, pvtkeylen);
	bf_encrypt(&bc, ky, kylen, ke, kelen);
	switch (data_encryption_type) {
		case 0:	/* no encryption */
			break;
		case BLOWFISH_ENCRYPTION:
			res = bf_setkey(&bc, ky, kylen);
			break;
		default:
			logexit(PIKTLOG, LOG_ERR, ERRID_DATA_ENCRYPTION_TYPE_INVALID, __FILE__, __LINE__, fnnam, FALSE, "invalid data_encryption_type %d", data_encryption_type);
			break;
	}
	return res;
}

/* generate the auth/session key; generate a random string combining various
   elements, including more than one authrand() call, then return the MD5
   checksum of the generated string;

   if authrand() simply returns the value of rand(), and since there are
   now eight authrand() calls in genkey(), we assert that there are no fewer
   than (RAND_MAX)^8 possible genkey() random strings; the insertion of
   semi-random elements in between authrand() calls only adds to the
   possibilities and doesn't detract from the overall randomness; these
   insertions also protect against a cracker somehow guessing our current
   sequence of authrand() numbers, for they introduce separate and
   independent sources of randomness/semi-randomness to the mix

   possible improvements to this algorithm:  substituting a better, more
   truly random authrand(); adding more elements to the rndstr mix; jumbling
   the rndstr before passing to checksum() (by, for example, stepping thru
   rndstr, char by char, and, with a 50% (or other) probability, switching
   adjoining chars); etc.
 */
char *
genkey (void)
{
	static char fnnam[] = "genkey";
	char rndstr[1024];	/* 1024 should be amply long; if not, up it */
	char *ky;

	snprintf(rndstr, 1024, "%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d",
	                       authrand(),
	                       authuid,           /* PIKT.conf uid */
	                       authrand(),
	                       prgpid,            /* current process id */
	                       authrand(),
	                       (int)time_start,   /* time of program start */
	                       authrand(),
	                       RPCPRGNUM,         /* from /etc/rpc */
	                       authrand(),
	                       (int)time((time_t*)0),  /* current time */
	                       authrand(),
	                       prgppid,           /* parent process id */
	                       authrand(),
	                       authgid,           /* PIKT.conf gid */
	                       authrand());
	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "rndstr[] is %s", rndstr);
	ky = checksumstr(5, rndstr);
	return ky;
}

/* return a random integer; todo: employ a better random number generator
   than simple rand(); note that srand() is called in init() (in init.c)
 */
int
authrand (void)
{
	int r;

	r = rand();
	return r;
}

/* report if a file exists
 */
int *
LIST_FILE_1_SVC (rfil *rf, struct svc_req *rqstp)
{
	static char fnnam[] = "LIST_FILE_SVC";
	char *dir;
	static char *fn;
	struct stat fbuf;
	static int res;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering LIST_FILE_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((res = chkrequest(rf->keysum.txttype_val, (ritm *)NULL, rf, rqstp, "list_file")) != 0)
		return &res;

	dir = getlocdir(rf->dir);
	fn = (char *) mymalloc((strlen(dir)+strlen(rf->file)+strlen(rf->ext)+2) * sizeof(char), fnnam);
	sprintf(fn, "%s/%s%s", dir, rf->file, strcmp(rf->ext, "NIL") ? rf->ext : nullstring);
	res = stat(fn, &fbuf);	/* stat returns 0 on success,
				   -1 (ERROR) on failure
				 */
	free(dir);
	free(fn);
	return &res;
}

/* install a file
 */
int *
INSTALL_FILE_1_SVC (rfil *rf, struct svc_req *rqstp)
{
	static char fnnam[] = "INSTALL_FILE_SVC";
	char *dir;
	static char *fn;
	char cmd[MAXCMD+1];
	FILE *fp;
	char *filtext;
	int fillen;
	char *keysan;
	mode_t mode;
	uid_t filuid;
	gid_t filgid;
	static int res;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering INSTALL_FILE_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((res = chkrequest(rf->keysum.txttype_val, (ritm *)NULL, rf, rqstp, "install_file")) != 0)
		return &res;

	dir = getlocdir(rf->dir);
	fn = (char *) mymalloc((strlen(dir)+strlen(rf->file)+strlen(rf->ext)+2) * sizeof(char), fnnam);
	sprintf(fn, "%s/%s%s", dir, rf->file, strcmp(rf->ext, "NIL") ? rf->ext : nullstring);
	if (rf->backup) {
		sprintf(cmd, "%s -p %s %s.%s", cpcmd, fn, fn, bakext);
		system(cmd);
	}
	if ((fp = fopen(fn, "w")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", fn);
		free(dir);
		free(fn);
		res = ERROR;
		return &res;
	}
	switch (data_encryption_type) {
		case 0:	/* no encryption */
			filtext = rf->text.txttype_val;
			fillen = rf->text.txttype_len;
			break;
		case BLOWFISH_ENCRYPTION:
			if (endian_master != endian)
				fix_endian(rf->text.txttype_val, rf->text.txttype_len);
			bf_decrypt(&bc, rf->text.txttype_val, rf->text.txttype_len, &filtext, &fillen);
			break;
		default:
			logexit(PIKTLOG, LOG_ERR, ERRID_DATA_ENCRYPTION_TYPE_INVALID, __FILE__, __LINE__, fnnam, FALSE, "invalid data_encryption_type %d", data_encryption_type);
			break;
	}
	if (isdebug) {
		keysan = sanitize(key, keylen);
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "key (keysan) is %s", /* key */ keysan);
		free(keysan);
	}
	if (fillen != rf->len) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_SIZE_MISMATCH, __FILE__, __LINE__, fnnam, FALSE, "unencrypted file size doesn't equal intended file size", fn);
		free(dir);
		free(fn);
		res = ERROR;
		return &res;
	}
	res = fwrite(filtext, sizeof(char), rf->len, fp);
	switch (data_encryption_type) {
		case BLOWFISH_ENCRYPTION:
			free(filtext);
			break;
		case 0:	/* no encryption */
		default:
			break;
	}
	if (res < rf->len) {
		if (res == 0)
			res = errno;
		else
			res = ERROR;
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_WRITE_ERROR, __FILE__, __LINE__, fnnam, res == ERROR ? 0 : 1, "fwrite() error for %s", fn);
		if (fclose(fp)) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", fn);
			free(dir);
			free(fn);
			res = ERROR;
			return &res;
		}
		free(dir);
		free(fn);
		return &res;
	}
	res = fclose(fp);
	if (res != 0) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", fn);
		free(dir);
		free(fn);
		res = ERROR;
		return &res;
	}
	mode = (mode_t) strtol(rf->mode, NULL, 8);	/* octal */
	if ((res = chmod(fn, mode)) != 0) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_CHMOD_ERROR, __FILE__, __LINE__, fnnam, TRUE, "chmod() error for %s", fn);
		free(dir);
		free(fn);
		res = errno;
		return &res;
	}
	filuid = (uid_t) atoi(rf->filuid);
	filgid = (gid_t) atoi(rf->filgid);
	if ((res = chown(fn, filuid, filgid)) != 0) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_CHOWN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "chown() error for %s", fn);
		free(dir);
		free(fn);
		res = errno;
		return &res;
	}
	free(dir);
	free(fn);
	return &res;
}

/* checksum a file
 */
cksumres *
CKSUM_FILE_1_SVC (rfil *rf, struct svc_req *rqstp)
{
	static char fnnam[] = "CKSUM_FILE_SVC";
	char *dir;
	static char *fn;
	struct stat fbuf;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering CKSUM_FILE_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((csres.result = chkrequest(rf->keysum.txttype_val, (ritm *)NULL, rf, rqstp, "cksum_file")) != 0)
		return &csres;

	dir = getlocdir(rf->dir);
	fn = (char *) mymalloc((strlen(dir)+strlen(rf->file)+strlen(rf->ext)+2) * sizeof(char), fnnam);
	sprintf(fn, "%s/%s%s", dir, rf->file, strcmp(rf->ext, "NIL") ? rf->ext : nullstring);
	/* free the memory allocated to the previous cksum string, if any */
	FREEVOID(csres.cksumres_u.cksum);
	if (stat(fn, &fbuf) == 0) {
		/* rf->len is where we indicate the cksum level */
		csres.cksumres_u.cksum = checksumfil((int) rf->len, fn);
		if (isverboselog || isdebug)
			logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "csres.cksumres_u.cksum is %s", csres.cksumres_u.cksum);
		if (!strcmp(csres.cksumres_u.cksum, errstr)) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CHECKSUM_ERROR, __FILE__, __LINE__, fnnam, FALSE, "checksumfil() error for %s", fn);
			csres.result = ERROR;
		} else {
			csres.result = 0;
		}
	} else {	/* file not found, so can't checksum it */
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_NOT_FOUND, __FILE__, __LINE__, fnnam, FALSE, "%s not found", fn);
		csres.result = NONE;
	}
	free(dir);
	free(fn);
	return &csres;
}

/* delete a file
 */
int *
DELETE_FILE_1_SVC (rfil *rf, struct svc_req *rqstp)
{
	static char fnnam[] = "DELETE_FILE_SVC";
	char *dir;
	static char *fn;
	char cmd[MAXCMD+1];
	struct stat fbuf;
	static int res;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering DELETE_FILE_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((res = chkrequest(rf->keysum.txttype_val, (ritm *)NULL, rf, rqstp, "delete_file")) != 0)
		return &res;

	dir = getlocdir(rf->dir);
	fn = (char *) mymalloc((strlen(dir)+strlen(rf->file)+strlen(rf->ext)+2) * sizeof(char), fnnam);
	sprintf(fn, "%s/%s%s", dir, rf->file, strcmp(rf->ext, "NIL") ? rf->ext : nullstring);
	if (rf->backup) {
		sprintf(cmd, "%s -p %s %s.%s", cpcmd, fn, fn, bakext);
		system(cmd);
	}
	if (stat(fn, &fbuf) == 0) {
		res = remove(fn);	/* remove returns 0 on success */
		if (res == ERROR) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "remove() error for %s", fn);
			res = errno;
		}
	} else	/* file not found, so signify success in removing it */
		res = 0;
	free(dir);
	free(fn);
	return &res;
}

/* restore a file from backup
 */
int *
RESTORE_FILE_1_SVC (rfil *rf, struct svc_req *rqstp)
{
	static char fnnam[] = "RESTORE_FILE_SVC";
	char *dir;
	static char *fn;
	static char *fb;
	struct stat fbuf;
	static int res;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering RESTORE_FILE_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((res = chkrequest(rf->keysum.txttype_val, (ritm *)NULL, rf, rqstp, "restore_file")) != 0)
		return &res;

	dir = getlocdir(rf->dir);
	fn = (char *) mymalloc((strlen(dir)+strlen(rf->file)+strlen(rf->ext)+2) * sizeof(char), fnnam);
	sprintf(fn, "%s/%s%s", dir, rf->file, strcmp(rf->ext, "NIL") ? rf->ext : nullstring);
	fb = (char *) mymalloc((strlen(fn)+strlen(bakext)+2) * sizeof(char), fnnam);
	sprintf(fb, "%s.%s", fn, bakext);
	if (stat(fb, &fbuf) == 0) {
		res = rename(fb, fn);	/* rename returns 0 on success */
		if (res == ERROR) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_RENAME_ERROR, __FILE__, __LINE__, fnnam, TRUE, "rename() error for %s and %s", fn);
			res = errno;
		}
	} else	{ /* backup file not found, so signify failure in restoring it */
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_NOT_FOUND, __FILE__, __LINE__, fnnam, FALSE, "%s not found", fb);
		res = errno;
	}
	free(dir);
	free(fn);
	free(fb);
	return &res;
}

/* for both enablement and disablement, for the given alert, remove its
   line from the piktd.conf file (if present); for enablement, add a new
   line to that file for the given alert
 */
int *
EDIT_PIKTD_CONF_1_SVC (ritm *an, struct svc_req *rqstp)
{
	static char fnnam[] = "EDIT_PIKTD_CONF_SVC";
	FILE *pcfp, *tpcfp;
	char line[MAXLINE+1];
	char target[MAXLINE+1];
	static int res;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering EDIT_PIKTD_CONF_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((res = chkrequest(an->keysum.txttype_val, an, (rfil *)NULL, rqstp, "edit_piktd_conf")) != 0)
		return &res;

	if ((pcfp = fopen(piktdconf, "r")) == NULL) {
		logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", piktdconf);
		/* this might be a brand new installation, hence no prior
		   piktd.conf file
		 */
		/*
		res = ERROR;
		return &res;
		 */
	}
	if ((tpcfp = fopen(piktdconftmp, "w")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", piktdconftmp);
		if (pcfp) {
			if (fclose(pcfp))
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", piktdconf);
		}
		res = ERROR;
		return &res;
	}
	if (pcfp) {
		sprintf(target, "+A %s\n", an->name);
		while (fgets(line, MAXLINE+1, pcfp) != NULL) {
			if (strstr(line, target) && line[0] != '#')
				continue;
			fputs(line, tpcfp);
		}
	}
	if ((an->state == ADD) || (an->aux && *(an->aux)))
		fprintf(tpcfp, "%s\n", an->aux);
	if (fclose(tpcfp)) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", piktdconftmp);
		res = ERROR;
		return &res;
	}
	if (pcfp) {
		if (fclose(pcfp)) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", piktdconf);
			res = ERROR;
			return &res;
		}
		if ((res = remove(piktdconf)) == ERROR) {	/* remove returns 0 on success */
			logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "remove() error for %s", piktdconf);
			return &res;
		}
	}
	if ((res = rename(piktdconftmp, piktdconf)) == ERROR) {	/* 0 success, -1 failure */
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_RENAME_ERROR, __FILE__, __LINE__, fnnam, TRUE, "rename() error for %s and %s", piktdconftmp, piktdconf);
		return &res;
	}
	if (chmod(piktdconf, (mode_t)(S_IRUSR|S_IWUSR|S_IRGRP)) != 0)
		logreturn(PIKTLOG, LOG_ERR, ERRID_CHMOD_ERROR, __FILE__, __LINE__, fnnam, TRUE, "chmod() error for %s", piktdconf);
	res = 0;
	return &res;
}

/* execute a command with no wait
 */
int *
EXECUTE_COMMAND_1_SVC (ritm *cn, struct svc_req *rqstp)
{
	static char fnnam[] = "EXECUTE_COMMAND_SVC";
	static char cmd[MAXLINE+1];
	pid_t pid;
	static int res;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering EXECUTE_COMMAND_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((res = chkrequest(cn->keysum.txttype_val, cn, (rfil *)NULL, rqstp, cn->name)) != 0)
		return &res;

	if (cn->aux && *(cn->aux)) {
		if (strlen(cn->aux) > MAXLINE) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_COMMAND_STRING_LEN_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "command string exceeds MAXLINE (%d) chars", MAXLINE);
			res = ERROR;
			return &res;
		} else
			strcpy(cmd, cn->aux);
	} else {
		logreturn(PIKTLOG, LOG_ERR, ERRID_COMMAND_STRING_MISSING, __FILE__, __LINE__, fnnam, FALSE, "missing command string");
		res = ERROR;
		return &res;
	}

	switch (pid = fork()) {
		case -1:	/* error */
			logreturn(PIKTLOG, LOG_ERR, ERRID_FORK_ERROR, __FILE__, __LINE__, fnnam, TRUE,  "fork() error");
			break;
		case 0:		/* child */
			logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "%s", cmd);
			execl(SHELL, "sh", "-c", cmd, (char *) NULL);
			_exit(127);
			break;
		default:	/* parent */
			break;
	}
	res = 0;
	return &res;
}

/* execute a command with wait
 */
execres *
EXECUTE_COMMANDWAIT_1_SVC (ritm *cn, struct svc_req *rqstp)
{
	static char fnnam[] = "EXECUTE_COMMANDWAIT_SVC";
	static char cmd[MAXLINE+1];
	char *cp;
	char *dir;
	FILE *cmdproc;
	char *execcmdout;
	unsigned int execcmdoutlen;
	char *keysan;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering EXECUTE_COMMANDWAIT_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((xres.result = chkrequest(cn->keysum.txttype_val, cn, (rfil *)NULL, rqstp, cn->name)) != 0)
		return &xres;

	if (cn->aux && *(cn->aux)) {
		if (strlen(cn->aux)+5 > MAXLINE) {	/* +5 for "2>&1" */
			logreturn(PIKTLOG, LOG_ERR, ERRID_COMMAND_STRING_LEN_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "command string exceeds MAXLINE (%d) chars", MAXLINE);
			xres.result = ERROR;
			return &xres;
		} else {
			/* special handling if this is a fetch_file()
			   from the piktmaster
			 */
			if (! strncmp(cn->name, "fetch_file", 10)) {
				if ((cp = strstr(cn->aux, ALTDIR))) {
					dir = getlocdir(ALTDIR);
					sprintf(cmd, "%s %s%s", catcmd, dir, cp+DIRLEN);
					free(dir);
				} else if ((cp = strstr(cn->aux, PRGDIR))) {
					dir = getlocdir(PRGDIR);
					sprintf(cmd, "%s %s%s", catcmd, dir, cp+DIRLEN);
					free(dir);
				} else if ((cp = strstr(cn->aux, OBJDIR))) {
					dir = getlocdir(OBJDIR);
					sprintf(cmd, "%s %s%s", catcmd, dir, cp+DIRLEN);
					free(dir);
				} else {
					sprintf(cmd, "%s", cn->aux);
				}
			} else {
				/* check to see if "2>&1" already part of
				   cn->aux?
				 */
				sprintf(cmd, "%s 2>&1", cn->aux);
			}
		}
	} else {
		logreturn(PIKTLOG, LOG_ERR, ERRID_COMMAND_STRING_MISSING, __FILE__, __LINE__, fnnam, FALSE, "missing command string");
		xres.result = ERROR;
		return &xres;
	}
	/* free execcmdout before executing the command */
	free_execcmdout();
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "%s", cmd);
	if ((cmdproc = popen(cmd, "r")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_PROC_OPEN_ERROR, __FILE__, __LINE__, fnnam, FALSE, "popen() error");
		xres.result = ERROR;
		return &xres;
	}
	execcmdout = (char *) mycalloc((size_t)(execcmdout_limit+1), sizeof(char), fnnam);
	if ((fread(execcmdout, sizeof(char), execcmdout_limit, cmdproc) == 0) && ferror(cmdproc)) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_READ_ERROR, __FILE__, __LINE__, fnnam, FALSE, "fread() error for %s", cmd);
		xres.result = ERROR;
	} else {
		execcmdoutlen = strlen(execcmdout);
		switch (data_encryption_type) {
			case 0:	/* no encryption */
				xres.execres_u.execcmdout.txttype_val = execcmdout;
				xres.execres_u.execcmdout.txttype_len = execcmdoutlen;
				break;
			case BLOWFISH_ENCRYPTION:
				bf_encrypt(&bc, execcmdout, execcmdoutlen, &xres.execres_u.execcmdout.txttype_val, &xres.execres_u.execcmdout.txttype_len);
				if (endian_master != endian)
					fix_endian(xres.execres_u.execcmdout.txttype_val, xres.execres_u.execcmdout.txttype_len);
				break;
			default:
				logexit(PIKTLOG, LOG_ERR, ERRID_DATA_ENCRYPTION_TYPE_INVALID, __FILE__, __LINE__, fnnam, FALSE, "invalid data_encryption_type %d", data_encryption_type);
				break;
		}
		if (isdebug) {
			keysan = sanitize(key, keylen);
			logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "key (keysan) is %s", /* key */ keysan);
			free(keysan);
		}
		xres.result = 0;
	}
	switch (data_encryption_type) {
		case BLOWFISH_ENCRYPTION:
			free(execcmdout);
			break;
		case 0:	/* no encryption */
		default:
			break;
	}
	pclose(cmdproc);
	return &xres;
}

/* garbage collect the memory hunk previously allocated to temporary storage
   of exec_commandwait() output
 */
void
free_execcmdout (void)
{
	FREEVOID(xres.execres_u.execcmdout.txttype_val);
	return;
}

/* for the given alert, exec its corresponding confline (passed from the
   master, not read from the client piktd.conf)
 */
int *
EXECUTE_ALERT_1_SVC (ritm *an, struct svc_req *rqstp)
{
	static char fnnam[] = "EXECUTE_ALERT_SVC";
	static char *fn;
	struct stat fbuf;
	static char cmd[MAXLINE+1];
	register char *cp;
	register int i;
	int spcnt;
	pid_t pid;
	static int res;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering EXECUTE_ALERT_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((res = chkrequest(an->keysum.txttype_val, an, (rfil *)NULL, rqstp, "execute_alert")) != 0)
		return &res;

	fn = (char *) mymalloc((strlen(altdir)+strlen(an->name)+6) * sizeof(char), fnnam);
	sprintf(fn, "%s/%s.alt", altdir, an->name);
	if (stat(fn, &fbuf)) {	/* stat returns 0 on success,
				   -1 (ERROR) on failure
				 */
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_NOT_FOUND, __FILE__, __LINE__, fnnam, FALSE, "%s/%s.alt not found", altdir, an->name);
		free(fn);
		res = ERROR;
		return &res;
	}
	free(fn);

	if (an->aux && *(an->aux)) {
		cp = an->aux;
		spcnt = 0;
		/* count spaces until we get just past the 6 timing specs */
		while (*cp && (spcnt < 6)) {
			if (isspace((int)*cp))
				spcnt++;
			cp++;
		}
		if (! *cp || (spcnt < 6)) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_COMMAND_STRING_INVALID, __FILE__, __LINE__, fnnam, FALSE, "bad command string");
			res = ERROR;
			return &res;
		} else {
			i = 0;
			while (! isspace((int)*cp)) {
				cmd[i++] = *(cp++);
			}
			if (istest) {
				cmd[i++] = ' ';
				cmd[i++] = '-';
				cmd[i++] = 'T';
			}
			while (*cp) {
				cmd[i++] = *(cp++);
			}
			cmd[i] = '\0';
			strcat(cmd, " >/dev/null 2>&1");
		}
	} else {
		logreturn(PIKTLOG, LOG_ERR, ERRID_COMMAND_STRING_MISSING, __FILE__, __LINE__, fnnam, FALSE, "missing command string");
		res = ERROR;
		return &res;
	}

	switch (pid = fork()) {
		case -1:	/* error */
			logreturn(PIKTLOG, LOG_ERR, ERRID_FORK_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fork() error");
			break;
		case 0:		/* child */
			logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "%s", cmd);
			execl(SHELL, "sh", "-c", cmd, (char *) NULL);
			_exit(127);
			break;
		default:	/* parent */
			break;
	}
	res = 0;
	return &res;
}

/* kill the piktd daemon
 */
int *
KILL_PIKTD_1_SVC (rqst *rq, struct svc_req *rqstp)
{
	static char fnnam[] = "KILL_PIKTD_SVC";
	pid_t oldpid, newpid;
	struct stat fbuf;
	char cmd[MAXCMD+1];
	char line[MAXLINE+1];
	FILE *fc;
	static int res;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering KILL_PIKTD_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((res = chkrequest(rq->keysum.txttype_val, (ritm *)NULL, (rfil *)NULL, rqstp, "kill_piktd")) != 0)
		return &res;

	if ((oldpid = procid("piktd", PROCID)) == ERROR) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_MULTIPLE_PIKTD, __FILE__, __LINE__, fnnam, FALSE, "unable to kill piktd: multiple piktd's running");
		res = ERROR;
		return &res;
	} else if (oldpid == NONE) {
		res = NONE;
		return &res;
	}
	if (kill(oldpid, SIGKILL) != 0) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_KILL_ERROR, __FILE__, __LINE__, fnnam, TRUE, "failure of kill command");
		res = ERROR;
		return &res;
	}
	if ((newpid = procid("piktd", PROCID)) != NIL) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_KILL_ERROR, __FILE__, __LINE__, fnnam, FALSE, "unable to kill piktd (pid %d)", newpid);
		res = ERROR;
		return &res;
	} else
		logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "piktd (pid %d) killed", oldpid);
	/* killing piktd, and pikt runs, may yield a core file */
	if (lstat("/core", &fbuf) == 0) {
		sprintf(cmd, "%s /core", filecmd);
		if ((fc = popen(cmd, "r")) == NULL) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_PROC_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "popen() error");
		} else {
			if (fgets(line, MAXLINE+1, fc) == NULL) {
				if (ferror(fc))
					logreturn(PIKTLOG, LOG_ERR, ERRID_FGETS_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fgets() error for %s", cmd);
			}
			pclose(fc);
			logreturn(PIKTLOG, LOG_WARNING, ERRID_CORE_FILE_FOUND, __FILE__, __LINE__, fnnam, FALSE, "found /core file: %s", line);
		}
		if (remove("/core") == 0)	/* remove returns 0 on success */
			logreturn(PIKTLOG, LOG_WARNING, ERRID_CORE_FILE_REMOVED, __FILE__, __LINE__, fnnam, FALSE, "rm'ed /core file");
	}
	/* remove the piktd lock file */
	if (lstat(piktdlock, &fbuf) == 0) {
		if (remove(piktdlock) != 0)	/* remove returns 0 on success */
			logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "unable to remove the %s file", piktdlock);
	}
	/* remove the piktd alerts file */
	if (lstat(piktdalt, &fbuf) == 0) {
		if (remove(piktdalt) != 0)	/* remove returns 0 on success */
			logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "unable to remove the %s file", piktdalt);
	}
	res = 0;
	return &res;
}

/* kill the piktc_svc daemon
 */
int *
KILL_PIKTC_SVC_1_SVC (rqst *rq, struct svc_req *rqstp)
{
	static char fnnam[] = "KILL_PIKTC_SVC_SVC";
	static int res;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering KILL_PIKTC_SVC_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((res = chkrequest(rq->keysum.txttype_val, (ritm *)NULL, (rfil *)NULL, rqstp, "kill_piktc_svc")) != 0) {
		return &res;
	}

	/* unregister the service */
	svc_unregister(PIKTCPROG, PIKTCVERS);
	/* remove the lock */
	if (rmlocks(TRUE, TRUE) == ERROR) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, FALSE, "rmlocks() error");
		res = ERROR;
		return &res;
	}
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "daemon killed");
	/* no &res returned to caller */
	exit(0);
}

/* restart the piktd daemon
 */
int *
RESTART_PIKTD_1_SVC (rqst *rq, struct svc_req *rqstp)
{
	static char fnnam[] = "RESTART_PIKTD_SVC";
	pid_t oldpid, newpid;
	pid_t pid;
	struct stat fbuf;
	static int res;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering RESTART_PIKTD_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((res = chkrequest(rq->keysum.txttype_val, (ritm *)NULL, (rfil *)NULL, rqstp, "restart_piktd")) != 0)
		return &res;

	if ((oldpid = procid("piktd", PROCID)) == NIL) {
		/* if no piktd's running, remove any orphaned piktd lock
		   files
		 */
		if (lstat(piktdlock, &fbuf) == 0) {
			if (remove(piktdlock) != 0) {	/* remove returns 0 on success */
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "unable to restart piktd: unable to remove %s file", piktdlock);
				res = ERROR;
				return &res;
			}
		}
	} else if (oldpid == ERROR) { /* we have multiple piktd's running */
		logreturn(PIKTLOG, LOG_ERR, ERRID_MULTIPLE_PIKTD, __FILE__, __LINE__, fnnam, TRUE, "unable to restart piktd: multiple piktd's running");
		res = ERROR;
		return &res;
	} else { /* we have a running piktd */
		logreturn(PIKTLOG, LOG_ERR, ERRID_MULTIPLE_PIKTD, __FILE__, __LINE__, fnnam, TRUE, "unable to restart piktd: another piktd (pid %d) running", oldpid);
		res = ERROR;
		return &res;
	}
	switch (pid = fork()) {
		case -1:	/* error */
			logexit(PIKTLOG, LOG_ERR, ERRID_FORK_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fork() error");
			break;
		case 0:		/* child */
			execl(SHELL, "sh", "-c", piktd, (char *) NULL);
			_exit(127);
			break;
		default:	/* parent */
			/* rather than pause an arbitrary number of seconds
			   here, we could instead enter a while loop, repeating
			   a procid() of piktd until it returns a suitable
			   result (but under weird circumstances, might not
			   that risk having this parent piktc_svc hang?)
			 */
			sleep(fork_delay);
			break;
	}
	newpid = procid("piktd", PROCID);
	if ((newpid == ERROR) || (newpid == oldpid)) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_PIKTD_RESTART_ERROR, __FILE__, __LINE__, fnnam, FALSE, "failed to restart piktd");
		res = ERROR;
		return &res;
	}
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "piktd restart");
	res = 0;
	return &res;
}

/* restart the piktc_svc daemon
 */
int *
RESTART_PIKTC_SVC_1_SVC (rqst *rq, struct svc_req *rqstp)
{
	static char fnnam[] = "RESTART_PIKTC_SVC_SVC";
	static int res;
	pid_t pid;
	static char cmd[MAXLINE+1];

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering RESTART_PIKTC_SVC_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((res = chkrequest(rq->keysum.txttype_val, (ritm *)NULL, (rfil *)NULL, rqstp, "restart_piktc_svc")) != 0)
		return &res;

	/* form the cmd; we must be careful to restart with the same
	   options, especially when doing a 'make check'
	 */
	sprintf(cmd, "%s", piktc_svc);
	if (isdebug) {
		if (istest) {
			strcat(cmd, " -TG");
		} else {
			strcat(cmd, " -G");
		}
	} else if (istest) {
		strcat(cmd, " -T");
	}
	strcat(cmd, " &");
	/* unregister the service */
	svc_unregister(PIKTCPROG, PIKTCVERS);
	/* remove the lock */
	if (rmlocks(TRUE, TRUE) == ERROR) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, FALSE, "rmlocks() error");
		res = ERROR;
		return &res;
	}
	switch (pid = fork()) {
		case -1:	/* error */
			logreturn(PIKTLOG, LOG_ERR, ERRID_FORK_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fork() error");
			res = ERROR;
			return &res;
			break;
		case 0:		/* child */
			/* pause to give time for parent to die, else we
			   will be blocked by the presence of the parent
			   piktc_svc in the main() function's procid() check;
			   or we could enter a while loop here, repeating a
			   procid() of piktc_svc until it returns NIL (but
			   under weird circumstances, might not that risk
			   having this child piktc_svc hang, waiting and
			   waiting for its parent to die?)
			 */
			sleep(fork_delay);
			execl(SHELL, "sh", "-c", cmd, (char *) NULL);
			_exit(127);
			break;
		default:	/* parent */
			exit(exstat);
	}
	/* should never be reached */
	res = ERROR;
	return &res;
}

/* call procid() to determine the piktc_svc pid; for now, this is only
   called in the course of a restart_piktc_svc, so we do a chkrequest()
   of that service before supplying the pid
 */
int *
PIKTC_SVC_PID_1_SVC (rqst *rq, struct svc_req *rqstp)
{
	static char fnnam[] = "PIKTC_SVC_PID_SVC";
	static int res;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering PIKTC_SVC_PID_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((res = chkrequest(rq->keysum.txttype_val, (ritm *)NULL, (rfil *)NULL, rqstp, "piktc_svc_pid")) != 0)
		return &res;

	res = procid("piktc_svc", PROCID);
	return &res;
}

/* report the status of the given alert; an alert is inactive if: there is
   no .alt file, there is no line for that alert in piktd.conf, there is
   no active piktd, and/or the alert is absent from the piktd alerts file
 */
int *
SHOW_ALERT_STATUS_1_SVC (ritm *an, struct svc_req *rqstp)
{
	static char fnnam[] = "SHOW_ALERT_STATUS_SVC";
	FILE *pcfp, *rfp;
	static char *fn;
	char line[MAXLINE+1];
	char target[MAXLINE+1];
	struct stat fbuf;
	static int res;
	char *cp;

	if (isdebug) {
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "entering SHOW_ALERT_SVC()");
		sleep(DEBUG_TIME_TO_SLEEP);
	}

	if ((res = chkrequest(an->keysum.txttype_val, an, (rfil *)NULL, rqstp, "show_alert_status")) != 0)
		return &res;

	/* is there an .alt file? */
	fn = (char *) mymalloc((strlen(altdir)+strlen(an->name)+6) * sizeof(char), fnnam);
	sprintf(fn, "%s/%s.alt", altdir, an->name);
	if ((res = stat(fn, &fbuf)) == ERROR) {	/* 0 success, -1 failure */
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "stat() error for %s", fn);
		free(fn);
		return &res;
	}
	free(fn);

	/* found in piktd.conf? */
	if ((pcfp = fopen(piktdconf, "r")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", piktdconf);
		res = ERROR;
		return &res;
	}
	sprintf(target, "+A %s", an->name);
	res = ERROR;
	while (fgets(line, MAXLINE+1, pcfp) != NULL) {
		if (strstr(line, target) && line[0] != '#') {
			res = 0;
			break;
		}
	}
	if (fclose(pcfp)) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", piktdconf);
		res = ERROR;
		return &res;
	}
	if (res == ERROR) {
		logreturn(PIKTLOG, LOG_WARNING, ERRID_PIKTDCONF_ENTRY_MISSING, __FILE__, __LINE__, fnnam, FALSE, "%s entry not found in %s", an->name, piktdconf);
		return &res;
	}

	/* active piktd? */
	if ((res = procid("piktd", PROCID)) == ERROR) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_PIKTD_NOT_RUNNING, __FILE__, __LINE__, fnnam, FALSE, "active piktd not found");
		return &res;
	}

	/* registered in active piktd? */
	if ((rfp = fopen(piktdalt, "r")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", piktdalt);
		res = ERROR;
		return &res;
	}
	res = ERROR;
	while (fgets(line, MAXLINE+1, rfp) != NULL) {
		if ((cp = strchr(line, '\n')) != NULL)
			*cp = '\0';
		if (! strcmp(line, an->name)) {
			res = 0;
			break;
		}
	}
	if (fclose(rfp)) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", piktdalt);
		res = ERROR;
		return &res;
	}
	if (res == ERROR) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_NOT_FOUND, __FILE__, __LINE__, fnnam, FALSE, "%s not found in %s", an->name, piktdalt);
		return &res;
	}

	return &res;
}
